////////////////////////////////////////////////////////////////////////////////
//     
//    __                                ______
//   / /_______________  ____  ___     / ____/
//  / __/ ___/ ___/ __ \/ __ \/ _ \   /___ )  
// / /_(__  ) /__/ /_/ / /_/ /  __/  ____/ /  
// \__/____/\___/\____/ .___/\___/  /_____/   
//                   /_/                      
//     
/// \file cedrusbox.c
/// Definitions of cedrusbox functions.
/// \example cedrusbox01.c
/// \example cedrusbox02.c
/// \todo SV1 testen
////////////////////////////////////////////////////////////////////////////////


#include "../include/tscope5/cedrusbox.h"
#include "../include/tscope5/cedrusbox_internal.h"
#include "../include/tscope5/system_internal.h"


////////////////////////////////////////////////////////////////////////////////
/// @name Response registration
/// Cedrusboxes can be used as response devices. 
/// See timer.c for more information about response registration.
//@{
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
/// Define a cedrusbox button as a response button.
///
/// \param device Number of the cedrusbox (devices are numbered from 1).
/// \param button Number of the button (buttons are counted from 1).
///
/// \return The reponse number associated with the button.
///
/// Give a positive number if you want to monitor button press events,
/// a negative number if you want to monitor button release events.
////////////////////////////////////////////////////////////////////////////////
int ts5_define_cedrusbox_button(int device, int button)
{
	ts5_check_cedrusbox("ts5_define_cedrusbox_button");
	ts5_log(TS5_LOGLEVEL_3, "ts5_define_cedrusbox_button(%d,%d)\n", device, button);

	if (device<1 || device>_ts5_status.timer.num_cedrusboxes) {
		ts5_fatal("%s: %s %d, %s %d\n", "ts5_define_cedrusbox_button",
					"device argument is", device, "number of cedrusboxes is", 
					_ts5_status.timer.num_cedrusboxes);
	}

	if (button==0) {
		ts5_fatal("%s: %s\n", "ts5_define_cedrusbox_button",
				 	"button argument is 0, response buttons are numbered from 1");
	}

	if (abs(button)>_ts5_status.timer.cedrusbox[device-1].num_buttons) {
		ts5_fatal("%s: %s %d, %s %d is %d\n", "ts5_define_cedrusbox_button",
					"button argument is ", button,
					"number of cedrusbox buttons for device", device, 
					_ts5_status.timer.cedrusbox[device-1].num_buttons);
	}

	if (button>0 && _ts5_status.timer.cedrusbox[device-1]
					.button_press_defined[button-1] !=0) {
		ts5_fatal("%s: %s %d %s\n", "ts5_define_cedrusbox_button",
					"button press", button, "is already defined");
	}

	if (button<0 && _ts5_status.timer.cedrusbox[device-1]
					.button_release_defined[-button-1]!=0) {
		ts5_fatal("%s: %s %d %s\n", "ts5_define_cedrusbox_button",
				 	"button release", button, "is already defined");
	}

	_ts5_status.timer.cedrusbox_is_response_device = 1;
	_ts5_status.timer.num_defined_buttons++;
	_ts5_status.timer.num_active_buttons++;
	_ts5_status.timer.cedrusbox[device-1].num_defined_buttons++;
	_ts5_status.timer.cedrusbox[device-1].num_active_buttons++;

	if (button>0) {
		_ts5_status.timer.cedrusbox[device-1]
			.button_press_defined[button-1] =
			_ts5_status.timer.num_defined_buttons;

		_ts5_status.timer.cedrusbox[device-1]
			.button_press_active[button-1] =
			_ts5_status.timer.num_defined_buttons;

	}
	else {
		_ts5_status.timer.cedrusbox[device-1]
			.button_release_defined[-button-1] =
			_ts5_status.timer.num_defined_buttons;

		_ts5_status.timer.cedrusbox[device-1]
			.button_release_active[-button-1] =
			_ts5_status.timer.num_defined_buttons;
	}

	return _ts5_status.timer.num_defined_buttons;
}


////////////////////////////////////////////////////////////////////////////////
/// Get the number of cedrusboxes that are connected to the system.
///
/// \return The number of cedrusboxes that are connected to the system.
////////////////////////////////////////////////////////////////////////////////
int ts5_get_num_cedrusboxes()
{
	ts5_check_cedrusbox("ts5_get_num_cedrusboxes");
	ts5_log(TS5_LOGLEVEL_4, "ts5_get_num_cedrusboxes()\n");
	
	return _ts5_status.timer.num_cedrusboxes;	
}


////////////////////////////////////////////////////////////////////////////////
/// Get the number of buttons available on a cedrusbox.
///
/// \param device Number of the cedrusbox (devices are numbered from 1).
///
/// \return The number of buttons on the cedrusbox.
////////////////////////////////////////////////////////////////////////////////
int ts5_get_num_cedrusbox_buttons(int device)
{
	ts5_check_cedrusbox("ts5_get_num_cedrusbox_buttons");
	ts5_log(TS5_LOGLEVEL_4, "ts5_get_num_cedrusbox_buttons(%d)\n", device);
	
	if (device<1 || device>_ts5_status.timer.num_cedrusboxes) {
		ts5_fatal("%s: %s %d, %s %d\n", "ts5_get_num_cedrusbox_buttons",
					"device argument is", device, "number of cedrusboxes is", 
					_ts5_status.timer.num_cedrusboxes);
	}
	
	return _ts5_status.timer.cedrusbox[device-1].num_buttons;
}


////////////////////////////////////////////////////////////////////////////////
//@}
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
/// @name Trigger input
/// Cedrusboxes can be used as trigger input devices. 
/// See timer.c for more information about trigger input.
//@{
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
/// Define a cedrusbox as a trigger input device
///
/// \param device Number of the cedrusbox (devices are numbered from 1).
////////////////////////////////////////////////////////////////////////////////
void ts5_define_cedrusbox_trigger_input(int device)
{
	ts5_check_cedrusbox("ts5_define_cedrusbox_trigger_input");
	ts5_log(TS5_LOGLEVEL_3, "ts5_define_cedrusbox_trigger_input(%d)\n", device);

	if (device<1 || device>_ts5_status.timer.num_cedrusboxes) {
		ts5_fatal("%s: %s %d, %s %d\n", "ts5_define_cedrusbox_trigger_input",
					"device argument is", device, "number of cedrusboxes is", 
					_ts5_status.timer.num_cedrusboxes);
	}

	_ts5_status.timer.cedrusbox_is_trigger_device = 1;
	_ts5_status.timer.cedrusbox[device-1].is_trigger_input_device = 1;
}


////////////////////////////////////////////////////////////////////////////////
/// Turn on cedrusbox trigger input simulation.
///
/// \param device Number of the cedrusbox (devices are numbered from 1).
/// \param interval Interval in seconds between two triggers.
////////////////////////////////////////////////////////////////////////////////
void ts5_simulate_cedrusbox_trigger_input(int device, double interval)
{
	ts5_check_cedrusbox("ts5_simulate_cedrusbox_trigger_input");
	ts5_log(TS5_LOGLEVEL_3, "ts5_simulate_cedrusbox_trigger_input(%d,%f)\n", 
				device, interval);

	if (device<1 || device>_ts5_status.timer.num_cedrusboxes) {
		ts5_fatal("%s: %s %d, %s %d\n", "ts5_simulate_cedrusbox_trigger_input",
					"device argument is", device, "number of cedrusboxes is", 
					_ts5_status.timer.num_cedrusboxes);
	}
								
	_ts5_status.timer.cedrusbox[device-1].simulate_trigger_input = 1;
	
	_ts5_status.timer.cedrusbox[device-1].trigger_simulation_interval = 
		interval;
}

////////////////////////////////////////////////////////////////////////////////
//@}
////////////////////////////////////////////////////////////////////////////////

