////////////////////////////////////////////////////////////////////////////////
//     
//    __                                ______
//   / /_______________  ____  ___     / ____/
//  / __/ ___/ ___/ __ \/ __ \/ _ \   /___ )  
// / /_(__  ) /__/ /_/ / /_/ /  __/  ____/ /  
// \__/____/\___/\____/ .___/\___/  /_____/   
//                   /_/                      
//     
/// \file timer.c
/// Definitions of timer functions.
/// \example timer01.c
/// \example timer02.c
/// \example timer03.c
/// \example timer04.c
/// \example timer05.c
/// \example timer06.c
////////////////////////////////////////////////////////////////////////////////


#include "../include/tscope5/timer.h"
#include "../include/tscope5/timer_internal.h"
#include "../include/tscope5/system_internal.h"

#ifdef TS5_WINDOWS
	 #include <windows.h>

#elif defined TS5_MACOSX
	#include <unistd.h>
	#include <sys/resource.h>
	#include <sys/types.h>

#elif defined TS5_LINUX
	#include <unistd.h>
	#include <sys/time.h>
	#include <sys/resource.h>

#endif

////////////////////////////////////////////////////////////////////////////////
/// @name Program priority functions
/// You can increase your program's priority to get better timing accuracy.
//@{
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
/// Set the program's priority
/// \param priority Priority. 
///
/// \return The previous priority.
///
/// There are five possible settings:
/// - TS5_HIGH_PRIORITY 
/// - TS5_ABOVE_NORMAL_PRIORITY
/// - TS5_NORMAL_PRIORITY
/// - TS5_BELOW_NORMAL_PRIORITY
/// - TS5_LOW_PRIORITY. 
///
/// The default priority is TS5_NORMAL_PRIORITY.
////////////////////////////////////////////////////////////////////////////////
int ts5_set_priority(int priority)
{
	ts5_check_timer("ts5_set_priority");
	ts5_log(TS5_LOGLEVEL_4, "ts5_set_priority(%d)\n", priority);

	int oldval = _ts5_status.timer.priority;


	#ifdef TS5_WINDOWS
		HANDLE h;
	 	h = GetCurrentProcess();
		
		DWORD priorityclass;

		if (priority == TS5_HIGH_PRIORITY) {
			priorityclass = HIGH_PRIORITY_CLASS;
		}
		else if (priority == TS5_ABOVE_NORMAL_PRIORITY) {
			priorityclass = ABOVE_NORMAL_PRIORITY_CLASS;
		}
		else if (priority == TS5_NORMAL_PRIORITY) {
			priorityclass = NORMAL_PRIORITY_CLASS;				
		}
		else if (priority == TS5_BELOW_NORMAL_PRIORITY) {
			priorityclass = BELOW_NORMAL_PRIORITY_CLASS;
		}
		else if (priority == TS5_LOW_PRIORITY) {
			priorityclass = IDLE_PRIORITY_CLASS;				
		}
		else {
			ts5_fatal("ts5_set_priority: unknown priority requested\n");
		}
			
		if (!SetPriorityClass(h, priorityclass)) {
			ts5_fatal("ts5_set_priority: can't change program priority\n");			
		}
	 
	#elif defined TS5_MACOSX
	    if (geteuid() != 0) {
			ts5_fatal("%s: %s, %s\n", "ts5_set_priority", "not running as root", 
						"can't change program priority");			
	    }
		
		int priorityclass = 0;
		
		if (priority == TS5_HIGH_PRIORITY) {
			priorityclass = -20;
		}
		else if (priority == TS5_ABOVE_NORMAL_PRIORITY) {
			priorityclass = -10;
		}
		else if (priority == TS5_NORMAL_PRIORITY) {
			priorityclass = 0;				
		}
		else if (priority == TS5_BELOW_NORMAL_PRIORITY) {
			priorityclass = 10;
		}
		else if (priority == TS5_LOW_PRIORITY) {
			priorityclass = 20;				
		}
		else {
			ts5_fatal("ts5_set_priority: unknown priority requested\n");
		}
		
		if (setpriority(PRIO_PGRP, 0, priorityclass)) {
			ts5_fatal("ts5_set_priority: can't change program priority\n");			
		}
		

	#elif defined TS5_LINUX
	    if (geteuid() != 0) {
			ts5_fatal("%s: %s, %s\n", "ts5_set_priority", "not running as root", 
						"can't change program priority");			
	    }
		
		ts5_check_timer4("ts5_set_priority");
			
		int priorityclass;
		
		if (priority == TS5_HIGH_PRIORITY) {
			priorityclass = -20;
		}
		else if (priority == TS5_ABOVE_NORMAL_PRIORITY) {
			priorityclass = -10;
		}
		else if (priority == TS5_NORMAL_PRIORITY) {
			priorityclass = 0;				
		}
		else if (priority == TS5_BELOW_NORMAL_PRIORITY) {
			priorityclass = 10;
		}
		else if (priority == TS5_LOW_PRIORITY) {
			priorityclass = 19;				
		}

		else {
			ts5_fatal("ts5_set_priority: unknown priority requested\n");
		}
		
		if (setpriority(PRIO_PGRP, 0, priorityclass)) {
			ts5_fatal("ts5_set_priority: can't change program priority\n");			
		}

	#endif
	
	_ts5_status.timer.priority = priority;
	
	return oldval;
}


////////////////////////////////////////////////////////////////////////////////
/// Get the program's priority
///
/// \return The program's priority.
////////////////////////////////////////////////////////////////////////////////
int ts5_get_priority()
{
	ts5_check_timer("ts5_get_priority");
	ts5_log(TS5_LOGLEVEL_4, "ts5_get_priority()\n");

	return _ts5_status.timer.priority;
}


////////////////////////////////////////////////////////////////////////////////
//@}
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
/// @name General timing functions
/// The general timing functions are mainly used to control stimulus timing.
/// 
/// You can either wait for a given amount of time, 
/// or until some time in the future. 
//@{
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
/// Get the number of seconds since the program started.
///
/// \return The number of seconds since the program started 
/// (as a double precision float).
////////////////////////////////////////////////////////////////////////////////
double ts5_get_time()
{
	ts5_check_timer("ts5_get_time");
	ts5_log(TS5_LOGLEVEL_5, "ts5_get_time()\n");

	return al_get_time();
}


////////////////////////////////////////////////////////////////////////////////
/// Wait for a number of seconds.
////////////////////////////////////////////////////////////////////////////////
void ts5_wait(double waittime)
{
	ts5_check_timer("ts5_wait");
	ts5_log(TS5_LOGLEVEL_5, "ts5_wait(%f)\n", waittime);

	// only use up cpu during the last 20 ms.
	double starttime = al_get_time();

	if (waittime > 0.02) {
		while (al_get_time() < starttime + waittime - 0.02) {
			ts5_realtime_clock_nap();
		}
	}

	while (al_get_time() < starttime+waittime) {

	}
}


////////////////////////////////////////////////////////////////////////////////
/// Wait until a deadline.
////////////////////////////////////////////////////////////////////////////////
void ts5_wait_until(double deadline)
{
	ts5_check_timer("ts5_wait_until");
	ts5_log(TS5_LOGLEVEL_5, "ts5_wait_until(%f)\n", deadline);

	// only use up cpu during the last 20 ms.
	double starttime = al_get_time();

	if (deadline - starttime > 0.02) {
		while (al_get_time() < deadline - 0.02) {
			ts5_realtime_clock_nap();
		}
	}

	while (al_get_time() < deadline) {

	}
}


////////////////////////////////////////////////////////////////////////////////
//@}
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
/// @name Response registration
/// The response registration functions require you to define 
/// response buttons first using ts5_define_mouse_button(),
/// ts5_define_keyboard_button(), ts5_define_joystick_button(),
/// ts5_define_cedrusbox_button() or ts5_define_parport_button(). 
/// 
/// Each response button will get a response value. 
/// The first button that is defined will get number 1, etc.
///
/// Registration of responses is done asynchronously. 
/// All button presses that are defined will be buffered 
/// continuously during the program.
///
/// Waiting for a response to a stimulus will typically involve 
/// flushing all buffered responses that occurred before stimulus onset using
/// ts5_flush_responses() and then checking for new responses using
/// ts5_check_response(), ts5_wait_for_response(),
/// ts5_wait_for_response_timed() or ts5_wait_for_response_until().
///
/// The response definition can be altered during the program using
/// ts5_hide_response_button() or ts5_remove_response_buttons().
//@{
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
/// Temporarily deactivates a response button.
///
/// \param button The button that has to be deactivated.
///
/// \return The number of active buttons left.
///
/// This function also checks whether there are any active response left.
/// Aborts if necessary.
///
/// Reactivating all defined buttons is possible by calling
/// ts5_hide_response_button(0).
////////////////////////////////////////////////////////////////////////////////
int ts5_hide_response_button(int button)
{
	ts5_check_timer("ts5_hide_response_button");
	ts5_check_timer2("ts5_hide_response_button");
	ts5_log(TS5_LOGLEVEL_3, "ts5_hide_response_button(%d)\n", button);

	TS5_TIMER_STATUS *timer = &_ts5_status.timer;
	
	if (button<0 || button>timer->num_defined_buttons) {
		ts5_fatal("%s: %s %d, %s %d\n", "ts5_hide_response_button",
				 	"button argument is", button, 
					"number of defined buttons is",
					timer->num_defined_buttons);
	}

	// reactivate
	int found = 1;

	if (button==0) {

		int i, j;

		if (timer->voicekey_is_response_device) {

			for (i=0; i<timer->voicekey.num_buttons; i++) {

				timer->voicekey.button_press_active[i] =
					timer->voicekey.button_press_defined[i];

				timer->voicekey.button_release_active[i] =
					timer->voicekey.button_release_defined[i];
			}

			timer->voicekey.num_active_buttons =
				timer->voicekey.num_defined_buttons;
		}
		
		if (timer->mouse_is_response_device) {

			for (i=0; i<timer->mouse.num_buttons; i++) {

				timer->mouse.button_press_active[i] =
					timer->mouse.button_press_defined[i];

				timer->mouse.button_release_active[i] =
					timer->mouse.button_release_defined[i];
			}

			timer->mouse.num_active_buttons =
				timer->mouse.num_defined_buttons;
		}

		if (timer->keyboard_is_response_device) {

			for (i=0; i<timer->keyboard.num_buttons; i++) {

				timer->keyboard.button_press_active[i] =
					timer->keyboard.button_press_defined[i];

				timer->keyboard.button_release_active[i] =
					timer->keyboard.button_release_defined[i];
			}

			timer->keyboard.num_active_buttons =
				timer->keyboard.num_defined_buttons;
		}

		if (timer->joystick_is_response_device) {

			for (i=0; i<timer->num_joysticks; i++) {

				for (j=0; j<timer->joystick[i].num_buttons; j++) {

					timer->joystick[i].button_press_active[j] =
						timer->joystick[i].button_press_defined[j];

					timer->joystick[i].button_release_active[j] =
						timer->joystick[i].button_release_defined[j];
				}
				timer->joystick[i].num_active_buttons =
					timer->joystick[i].num_defined_buttons;
			}
		}

		if (timer->cedrusbox_is_response_device) {

			for (i=0; i<timer->num_cedrusboxes; i++) {

				for (j=0; j<timer->cedrusbox[i].num_buttons; j++) {

					timer->cedrusbox[i].button_press_active[j] =
						timer->cedrusbox[i].button_press_defined[j];

					timer->cedrusbox[i].button_release_active[j] =
						timer->cedrusbox[i].button_release_defined[j];
				}
				timer->cedrusbox[i].num_active_buttons =
					timer->cedrusbox[i].num_defined_buttons;
			}
		}
		
		if (timer->parport_is_response_device) {

			for (i=0; i<timer->num_parports; i++) {

				for (j=0; j<timer->parport[i].num_buttons; j++) {

					timer->parport[i].button_press_active[j] =
						timer->parport[i].button_press_defined[j];

					timer->parport[i].button_release_active[j] =
						timer->parport[i].button_release_defined[j];
				}
				timer->parport[i].num_active_buttons =
					timer->parport[i].num_defined_buttons;
			}
		}

		if (timer->serialport_is_response_device) {

			for (i=0; i<timer->num_serialports; i++) {

				for (j=0; j<timer->serialport[i].num_buttons; j++) {

					timer->serialport[i].button_press_active[j] =
						timer->serialport[i].button_press_defined[j];

					timer->serialport[i].button_release_active[j] =
						timer->serialport[i].button_release_defined[j];
				}
				timer->serialport[i].num_active_buttons =
					timer->serialport[i].num_defined_buttons;
			}
		}
				
		timer->num_active_buttons = 
			timer->num_defined_buttons;
	}
	else {
		int i, j;
		found=0;

		if (timer->voicekey_is_response_device) {

			for (i=0; i<timer->voicekey.num_buttons; i++) {

				if (button==timer->voicekey.button_press_active[i]) {
					timer->voicekey.button_press_active[i] = 0;
					found = 1;
				}
				else if (button==timer->voicekey.button_release_active[i]) {
					timer->voicekey.button_release_active[i] = 0;
					found = 1;
				}

				if (found) {
					timer->voicekey.num_active_buttons--;
					timer->num_active_buttons--;
					goto end;
				}
			}
		}

		if (timer->mouse_is_response_device) {

			for (i=0; i<timer->mouse.num_buttons; i++) {

				if (button==timer->mouse.button_press_active[i]) {
					timer->mouse.button_press_active[i] = 0;
					found = 1;
				}
				else if (button==timer->mouse.button_release_active[i]) {
					timer->mouse.button_release_active[i] = 0;
					found = 1;
				}

				if (found) {
					timer->mouse.num_active_buttons--;
					timer->num_active_buttons--;
					goto end;
				}
			}
		}

		if (timer->keyboard_is_response_device) {

			for (i=0; i<timer->keyboard.num_buttons; i++) {

				if (button==timer->keyboard.button_press_active[i]) {
					timer->keyboard.button_press_active[i] = 0;
					found = 1;
				}
				else if (button==timer->keyboard.button_release_active[i]) {
					timer->keyboard.button_release_active[i] = 0;
					found = 1;
				}

				if (found) {
					timer->keyboard.num_active_buttons--;
					timer->num_active_buttons--;
					goto end;
				}
			}
		}

		if (timer->joystick_is_response_device) {

			for (i=0; i<timer->num_joysticks; i++) {

				for (j=0; j<timer->joystick[i].num_buttons; j++) {

					if (button==timer->joystick[i].button_press_active[j]) {
						timer->joystick[i].button_press_active[j] = 0;
						found = 1;
					}
					else if (button==timer->joystick[i].button_release_active[j]) {
						timer->joystick[i].button_release_active[j] = 0;
						found = 1;
					}

					if (found) {
						timer->joystick[i].num_active_buttons--;
						timer->num_active_buttons--;
						goto end;
					}
				}
			}
		}

		if (timer->cedrusbox_is_response_device) {

			for (i=0; i<timer->num_cedrusboxes; i++) {

				for (j=0; j<timer->cedrusbox[i].num_buttons; j++) {

					if (button==timer->cedrusbox[i].button_press_active[j]) {
						timer->cedrusbox[i].button_press_active[j] = 0;
						found = 1;
					}
					else if (button==timer->cedrusbox[i].button_release_active[j]){
						timer->cedrusbox[i].button_release_active[j] = 0;
						found = 1;
					}

					if (found) {
						timer->cedrusbox[i].num_active_buttons--;
						timer->num_active_buttons--;
						goto end;
					}
				}
			}
		}
		
		if (timer->parport_is_response_device) {

			for (i=0; i<timer->num_parports; i++) {

				for (j=0; j<timer->parport[i].num_buttons; j++) {

					if (button==timer->parport[i].button_press_active[j]) {
						timer->parport[i].button_press_active[j] = 0;
						found = 1;
					}
					else if (button==timer->parport[i].button_release_active[j]){
						timer->parport[i].button_release_active[j] = 0;
						found = 1;
					}

					if (found) {
						timer->parport[i].num_active_buttons--;
						timer->num_active_buttons--;
						goto end;
					}
				}
			}
		}

		if (timer->serialport_is_response_device) {

			for (i=0; i<timer->num_serialports; i++) {

				for (j=0; j<timer->serialport[i].num_buttons; j++) {

					if (button==timer->serialport[i].button_press_active[j]) {
						timer->serialport[i].button_press_active[j] = 0;
						found = 1;
					}
					else if (button==timer->serialport[i].button_release_active[j]){
						timer->serialport[i].button_release_active[j] = 0;
						found = 1;
					}

					if (found) {
						timer->serialport[i].num_active_buttons--;
						timer->num_active_buttons--;
						goto end;
					}
				}
			}
		}

	}

	end:

	if (!found) {
		ts5_fatal("ts5_hide_response_button: response button %d not found\n", button);
	}

	if (timer->num_active_buttons<1) {
		ts5_fatal("%s: %s\n", "ts5_hide_response_button",
				 	"there are no active response buttons left");
	}

	return timer->num_active_buttons;
}


////////////////////////////////////////////////////////////////////////////////
/// Completely removes the response button definition.
////////////////////////////////////////////////////////////////////////////////
void ts5_remove_response_buttons()
{
	ts5_check_timer("ts5_remove_response_buttons");
	ts5_log(TS5_LOGLEVEL_3, "ts5_remove_response_buttons()\n");

	TS5_TIMER_STATUS *timer = &_ts5_status.timer;

	int i, j;

	// voicekey
	if (timer->voicekey_is_response_device) {

		for (i=0; i<timer->voicekey.num_buttons; i++) {

			timer->voicekey.button_press_defined[i] = 0;
			timer->voicekey.button_press_active[i] = 0;
			timer->voicekey.button_release_defined[i] = 0;
			timer->voicekey.button_release_active[i] = 0;
		}

		timer->voicekey_is_response_device = 0;
	}

	// mouse
	if (timer->mouse_is_response_device) {

		for (i=0; i<timer->mouse.num_buttons; i++) {

			timer->mouse.button_press_defined[i] = 0;
			timer->mouse.button_press_active[i] = 0;
			timer->mouse.button_release_defined[i] = 0;
			timer->mouse.button_release_active[i] = 0;
		}

		timer->mouse_is_response_device = 0;
	}

	// keyboard
	if (timer->keyboard_is_response_device) {

		for (i=0; i<timer->keyboard.num_buttons; i++) {

			timer->keyboard.button_press_defined[i] = 0;
			timer->keyboard.button_press_active[i] = 0;
			timer->keyboard.button_release_defined[i] = 0;
			timer->keyboard.button_release_active[i] = 0;
		}

		timer->keyboard_is_response_device = 0;
	}

	// joystick
	if (timer->joystick_is_response_device) {

		for (i=0; i<timer->num_joysticks; i++) {

			for (j=0; j<timer->joystick[i].num_buttons; j++) {

				timer->joystick[i].button_press_defined[j] = 0;
				timer->joystick[i].button_press_active[j] = 0;
				timer->joystick[i].button_release_defined[j] = 0;
				timer->joystick[i].button_release_active[j] = 0;
			}
		}

		timer->joystick_is_response_device = 0;
	}

	// joystick
	if (timer->joystick_is_response_device) {

		for (i=0; i<timer->num_joysticks; i++) {

			for (j=0; j<timer->joystick[i].num_buttons; j++) {

				timer->joystick[i].button_press_defined[j] = 0;
				timer->joystick[i].button_press_active[j] = 0;
				timer->joystick[i].button_release_defined[j] = 0;
				timer->joystick[i].button_release_active[j] = 0;
			}
		}

		timer->joystick_is_response_device = 0;
	}

	// cedrusbox
	if (timer->cedrusbox_is_response_device) {

		for (i=0; i<timer->num_cedrusboxes; i++) {

			for (j=0; j<timer->cedrusbox[i].num_buttons; j++) {

				timer->cedrusbox[i].button_press_defined[j] = 0;
				timer->cedrusbox[i].button_press_active[j] = 0;
				timer->cedrusbox[i].button_release_defined[j] = 0;
				timer->cedrusbox[i].button_release_active[j] = 0;
			}
		}
		timer->cedrusbox_is_response_device = 0;
	}

	timer->num_defined_buttons=0;
	timer->num_active_buttons=0;
}


////////////////////////////////////////////////////////////////////////////////
/// Flush the response queue.
////////////////////////////////////////////////////////////////////////////////
void ts5_flush_responses()
{
	ts5_check_timer("ts5_flush_responses");
	ts5_log(TS5_LOGLEVEL_5, "ts5_flush_responses()\n");

	al_flush_event_queue(_ts5_data.timer.response_queue);
}


////////////////////////////////////////////////////////////////////////////////
/// Check for a reponse.
///
/// \param resptime Variable that will store the response time.
/// \param timing_error Variable that will store the timing error 
/// (works only for cedrusboxes and parallel port boxes).
///
/// \return The button that was pressed 
/// (0 if no button was pressed within the deadline).
///
/// This function checks for a response and returns immediately.
////////////////////////////////////////////////////////////////////////////////
int ts5_check_response(double *resptime, double *timing_error)
{
	ts5_check_timer("ts5_check_response");
	ts5_check_timer2("ts5_check_response");
	ts5_log(TS5_LOGLEVEL_6, "ts5_check_response(%p,%p)\n", 
				resptime, timing_error);

	int resp=0;
	ALLEGRO_EVENT event;

	TS5_TIMER_STATUS *timer = &_ts5_status.timer;

	if (!al_get_next_event(_ts5_data.timer.response_queue, &event)) {

		if (resptime!=NULL) {
			*resptime = al_get_time();
		}
		if (timing_error!=NULL) {
			*timing_error = 0.0;
		}
	}
	else {

		if (resptime!=NULL) {
			*resptime = event.any.timestamp;
		}

		if (timing_error!=NULL) {
			*timing_error = 0.0;
		}

		int i;

		// voicekey (driver response codes are counted from 1)
		if (timer->voicekey_is_response_device) {

			if (timer->voicekey.num_active_buttons) {

				if (event.type==TS5_EVENT_VOICEKEY_BUTTON_DOWN) {
					resp = timer->voicekey
							.button_press_active[event.user.data2-1];
				}

				if (event.type==TS5_EVENT_VOICEKEY_BUTTON_UP) {
					resp = timer->voicekey
							.button_release_active[event.user.data2-1];
				}
				
				if (resp && resptime!=NULL) {
					*resptime = ((double)event.user.data3)/1000000.0;
				}
				
			}
		}

		// mouse (driver response codes are counted from 1)
		if (timer->mouse_is_response_device) {

			if (timer->mouse.num_active_buttons) {

				if (event.type==ALLEGRO_EVENT_MOUSE_BUTTON_DOWN) {

					resp = timer->mouse
							.button_press_active[event.mouse.button-1];
				}

				if (event.type==ALLEGRO_EVENT_MOUSE_BUTTON_UP) {
					resp = timer->mouse
							.button_release_active[event.mouse.button-1];
				}
			}
		}

		// keyboard (driver response codes are counted from 1)
		if (timer->keyboard_is_response_device) {

			if (timer->keyboard.num_active_buttons) {

				if (event.type==ALLEGRO_EVENT_KEY_DOWN) {
					resp = timer->keyboard
							.button_press_active[event.keyboard.keycode-1];
				}

				if (event.type==ALLEGRO_EVENT_KEY_UP) {
					resp = timer->keyboard
							.button_release_active[event.keyboard.keycode-1];
				}
			}
		}

		// joystick (driver response codes are counted from 0)
		if (timer->joystick_is_response_device) {

			for (i=0; i<timer->num_joysticks; i++) {

				if (timer->joystick[i].num_active_buttons) {

					if (event.type==ALLEGRO_EVENT_JOYSTICK_BUTTON_DOWN) {

						ALLEGRO_JOYSTICK *stick = al_get_joystick(i);

						if (event.joystick.id == stick) {
							resp = timer->joystick[i]
									.button_press_active[event.joystick.button];
						}
					}

					if (event.type==ALLEGRO_EVENT_JOYSTICK_BUTTON_UP) {

						ALLEGRO_JOYSTICK *stick = al_get_joystick(i);

						if (event.joystick.id == stick) {
							resp = timer->joystick[i]
									.button_release_active
									[event.joystick.button];
						}
					}
				}
			}
		}

		// cedrusbox (driver response codes are counted from 0)
		if (timer->cedrusbox_is_response_device) {

			for (i=0; i<timer->num_cedrusboxes; i++) {

				if (timer->cedrusbox[i].num_active_buttons) {

					if (event.type==TS5_EVENT_CEDRUSBOX_BUTTON_DOWN) {

						if (event.user.data1 == i) {
							resp = timer->cedrusbox[i]
									.button_press_active[event.user.data2];
						}
					}

					if (event.type==TS5_EVENT_CEDRUSBOX_BUTTON_UP) {

						if (event.user.data1 == i) {
							resp = timer->cedrusbox[i]
									.button_release_active[event.user.data2];
						}
					}

					if (resp && timing_error!=NULL) {
						*timing_error = ((double)event.user.data3)/1000000.0;
					}
				}
			}
		}

		// parallel port (driver response codes are counted from 1)
		if (timer->parport_is_response_device) {

			for (i=0; i<timer->num_parports; i++) {

				if (timer->parport[i].num_active_buttons) {

					if (event.type==TS5_EVENT_PARPORT_BUTTON_DOWN) {

						if (event.user.data1 == i) {
							resp = timer->parport[i]
									.button_press_active[event.user.data2-1];
						}
					}

					if (event.type==TS5_EVENT_PARPORT_BUTTON_UP) {

						if (event.user.data1 == i) {
							resp = timer->parport[i]
									.button_release_active[event.user.data2-1];
						}
					}

					if (resp && timing_error!=NULL) {
						*timing_error = ((double)event.user.data3)/1000000.0;
					}
				}
			}
		}

		// serial port (driver response codes are counted from 1)
		if (timer->serialport_is_response_device) {
			
			for (i=0; i<timer->num_serialports; i++) {

				if (timer->serialport[i].num_active_buttons) {

					if (event.type==TS5_EVENT_SERIALPORT_BUTTON_DOWN) {

						if (event.user.data1 == i) {
							resp = timer->serialport[i]
									.button_press_active[event.user.data2-1];
						}
					}

					if (event.type==TS5_EVENT_SERIALPORT_BUTTON_UP) {

						if (event.user.data1 == i) {
							resp = timer->serialport[i]
									.button_release_active[event.user.data2-1];
						}
					}

					if (resp && timing_error!=NULL) {
						*timing_error = ((double)event.user.data3)/1000000.0;
					}
				}
			}
		}
	}

	return resp;
}


////////////////////////////////////////////////////////////////////////////////
/// Wait for a response.
///
/// \param resptime Variable that will store the response time.
/// \param timing_error Variable that will store the timing error 
/// (works only for cedrusboxes and parallel port boxes).
///
/// \return The button that was pressed.
////////////////////////////////////////////////////////////////////////////////
int ts5_wait_for_response(double *resptime, double *timing_error)
{
	ts5_check_timer("ts5_wait_for_response");
	ts5_check_timer2("ts5_wait_for_response");
	ts5_log(TS5_LOGLEVEL_5, "ts5_wait_for_response(%p,%p)\n", 
				resptime, timing_error);

	int resp=0;

	while (!resp) {
		resp = ts5_check_response(resptime, timing_error);
	}

	return resp;
}


////////////////////////////////////////////////////////////////////////////////
/// Wait for a response for a given time.
///
/// \param resptime Variable that will store the response time.
/// \param timing_error Variable that will store the timing error
/// (works only for cedrusboxes and parallel port boxes).
/// \param maxtime Maximum time for a response.
///
/// \return The button that was pressed 
/// (0 if no button was pressed within the maximum time).
////////////////////////////////////////////////////////////////////////////////
int ts5_wait_for_response_timed(double *resptime, double *timing_error, 
		double maxtime)
{
	ts5_check_timer("ts5_wait_for_response_timed");
	ts5_check_timer2("ts5_wait_for_response_timed");
	ts5_log(TS5_LOGLEVEL_5, "ts5_wait_for_response_timed(%p,%p,%f)\n", 
				resptime, timing_error, maxtime);

	if (maxtime<0) {
		ts5_fatal("ts5_wait_for_response_timed: maxtime is negative (%f)\n", 
			maxtime);
	}

	int resp=0, deadline=0;
	double starttime = al_get_time();

	while (!resp && !deadline) {

		resp = ts5_check_response(resptime, timing_error);

		if (al_get_time()-starttime>=maxtime) {
			deadline=1;
		}
	}

	return resp;
}


////////////////////////////////////////////////////////////////////////////////
/// Wait for a response until a deadline.
///
/// \param resptime Variable that will store the response time.
/// \param timing_error Variable that will store the timing error 
/// (works only for cedrusboxes and parallel port boxes).
/// \param deadline Deadline for a response.
///
/// \return The button that was pressed 
/// (0 if no button was pressed within the deadline).
////////////////////////////////////////////////////////////////////////////////
int ts5_wait_for_response_until(double *resptime, double *timing_error, 
									double deadline)
{
	ts5_check_timer("ts5_wait_for_response_until");
	ts5_check_timer2("ts5_wait_for_response_until");
	ts5_log(TS5_LOGLEVEL_5, "ts5_wait_for_response_until(%p,%p,%f)\n", 
				resptime, timing_error, deadline);

	if (deadline<al_get_time()) {
		ts5_fatal("%s: %s (%f)\n", "ts5_wait_for_response_until", 
				 	"response deadline is in the past",
					deadline-al_get_time());
	}

	int resp=0, maxtime=0;

	while (!resp && !maxtime) {

		resp = ts5_check_response(resptime, timing_error);

		if (al_get_time()>=deadline) {
			maxtime=1;
		}
	}

	return resp;
}


////////////////////////////////////////////////////////////////////////////////
//@}
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
/// @name Trigger input/output
/// The trigger input functions require you to define trigger devices
/// using ts5_define_parport_trigger_input() or
/// ts5_define_serialport_trigger_input().
///
/// Registration of trigger input is done asynchronously. 
/// All triggers received on the parallel and/or serial port(s) will be
/// buffered continuously during the program.
///
/// Waiting for a trigger is similar to waiting for a reponse.
/// Old triggers that were not processed yet can be flushed using
/// ts5_flush_triggers().
/// One can either wait for a trigger indefinitely, for a given amount of time
/// or until some deadline has passed.
///
/// The value of the trigger is defined by the computer that sends the trigger.
///
/// During the development of the program it is not necessary to connect a
/// computer that sends the triggers. 
/// Trigger input can be simulated using ts5_simulate_parport_trigger_input() 
/// or ts5_simulate_serialport_trigger_input().
///
/// At the end of the program all triggers can be written to a logfile using
/// ts5_write_all_triggers().
///
/// Trigger output is simple, call ts5_send_parport_trigger() or
/// ts5_send_serialport_trigger()
/// each time you want to send a trigger.
//@{
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
/// Flush the trigger queue.
////////////////////////////////////////////////////////////////////////////////
void ts5_flush_triggers()
{
	ts5_check_timer("ts5_flush_triggers");
	ts5_log(TS5_LOGLEVEL_5, "ts5_flush_triggers()\n");

	al_flush_event_queue(_ts5_data.timer.trigger_queue);
}


////////////////////////////////////////////////////////////////////////////////
/// Check for a trigger.
///
/// \param trigtime Variable that will store the trigger time.
/// \param timing_error Variable that will store the timing error.
///
/// \return The value of the trigger
///
/// This function checks for a trigger and returns immediately.
////////////////////////////////////////////////////////////////////////////////
int ts5_check_trigger(double *trigtime, double *timing_error)
{
	ts5_check_timer("ts5_check_trigger");
	ts5_check_timer3("ts5_check_trigger");
	ts5_log(TS5_LOGLEVEL_6, "ts5_check_trigger(%p,%p)\n", 
				trigtime, timing_error);

	int trigger=0;
	ALLEGRO_EVENT event;

	TS5_TIMER_STATUS *timer = &_ts5_status.timer;

	if (trigtime!=NULL) {
		*trigtime = al_get_time();
	}
	if (timing_error!=NULL) {
		*timing_error = 0.0;
	}

	if (al_get_next_event(_ts5_data.timer.trigger_queue, &event)) {

		if(timer->cedrusbox_is_trigger_device) {

			int i;
			for (i=0; i<timer->num_cedrusboxes; i++) {

				if (timer->cedrusbox[i].is_trigger_input_device) {

					if (event.type==TS5_EVENT_CEDRUSBOX_TRIGGER 
						&& event.user.data1 == i) {

						if (trigtime!=NULL) {
							*trigtime = event.any.timestamp;
						}

						if (timing_error!=NULL) {
							*timing_error = 
								((double)event.user.data3)/1000000.0;
						}

						trigger = event.user.data2;
					}
				}
			}
		}

		if(timer->parport_is_trigger_device) {

			int i;
			for (i=0; i<timer->num_parports; i++) {

				if (timer->parport[i].is_trigger_input_device) {

					if (event.type==TS5_EVENT_PARPORT_TRIGGER 
						&& event.user.data1 == i) {

						if (trigtime!=NULL) {
							*trigtime = event.any.timestamp;
						}

						if (timing_error!=NULL) {
							*timing_error = 
								((double)event.user.data3)/1000000.0;
						}

						trigger = event.user.data2;
					}
				}
			}
		}

		if(timer->serialport_is_trigger_device) {

			int i;
			for (i=0; i<timer->num_serialports; i++) {

				if (timer->serialport[i].is_trigger_input_device) {

					if (event.type==TS5_EVENT_SERIALPORT_TRIGGER 
						&& event.user.data1 == i) {

						if (trigtime!=NULL) {
							*trigtime = event.any.timestamp;
						}

						if (timing_error!=NULL) {
							*timing_error = 
								((double)event.user.data3)/1000000.0;
						}

						trigger = event.user.data2;
					}
				}
			}
		}
	}

	return trigger;
}


////////////////////////////////////////////////////////////////////////////////
/// Wait for a trigger.
///
/// \param trigtime Variable that will store the trigger time.
/// \param timing_error Variable that will store the timing error
/// (works only for cedrusboxes and parallel port boxes).
///
/// \return The button that was pressed.
////////////////////////////////////////////////////////////////////////////////
int ts5_wait_for_trigger(double *trigtime, double *timing_error)
{
	ts5_check_timer("ts5_wait_for_trigger");
	ts5_check_timer3("ts5_wait_for_trigger");
	ts5_log(TS5_LOGLEVEL_5, "ts5_wait_for_trigger(%p,%p)\n", 
				trigtime, timing_error);

	int trigger=0;

	while (!trigger) {
		trigger = ts5_check_trigger(trigtime, timing_error);
	}

	return trigger;
}


////////////////////////////////////////////////////////////////////////////////
/// Wait for a trigger for a given time.
///
/// \param trigtime Variable that will store the trigger time.
/// \param timing_error Variable that will store the timing error 
/// (works only for cedrusboxes and parallel port boxes).
/// \param maxtime Maximum time for a trigger.
///
/// \return The button that was pressed
/// (0 if no button was pressed within the maximum time).
////////////////////////////////////////////////////////////////////////////////
int ts5_wait_for_trigger_timed(double *trigtime, double *timing_error, 
									double maxtime)
{
	ts5_check_timer("ts5_wait_for_trigger_timed");
	ts5_check_timer3("ts5_wait_for_trigger_timed");
	
	ts5_log(TS5_LOGLEVEL_5, "ts5_wait_for_trigger_timed(%p,%p,%f)\n", 
				trigtime, timing_error, maxtime);

	if (maxtime<0) {
		ts5_fatal("ts5_wait_for_trigger_timed: maxtime is negative (%f)\n", 
					maxtime);
	}

	int trigger=0, deadline=0;
	double starttime = al_get_time();

	while (!trigger && !deadline) {

		trigger = ts5_check_trigger(trigtime, timing_error);

		if (al_get_time()-starttime>=maxtime) {
			deadline=1;
		}
	}

	return trigger;
}


////////////////////////////////////////////////////////////////////////////////
/// Wait for a trigger until a deadline.
///
/// \param trigtime Variable that will store the trigger time.
/// \param timing_error Variable that will store the timing error
/// (works only for cedrusboxes and parallel port boxes).
/// \param deadline Deadline for a trigger.
///
/// \return The button that was pressed 
/// (0 if no button was pressed within the deadline).
////////////////////////////////////////////////////////////////////////////////
int ts5_wait_for_trigger_until(double *trigtime, double *timing_error, 
								double deadline)
{
	ts5_check_timer("ts5_wait_for_trigger_until");
	ts5_check_timer3("ts5_wait_for_trigger_until");
	ts5_log(TS5_LOGLEVEL_5, "ts5_wait_for_trigger_until(%p,%p,%f)\n", 
				trigtime, timing_error, deadline);

	if (deadline<al_get_time()) {
		ts5_fatal("%s: %s (%f)\n", "ts5_wait_for_trigger_until",
				 	"trigger deadline is in the past",
					deadline-al_get_time());
	}

	int trigger=0, maxtime=0;

	while (!trigger && !maxtime) {

		trigger = ts5_check_trigger(trigtime, timing_error);

		if (al_get_time()>=deadline) {
			maxtime=1;
		}
	}

	return trigger;
}


////////////////////////////////////////////////////////////////////////////////
/// Write all triggers to an output file
///
/// \param filename Path to the output file
////////////////////////////////////////////////////////////////////////////////
void ts5_write_all_triggers(char *filename)
{
	ts5_check_timer("ts5_write_all_triggers");
	ts5_check_timer3("ts5_write_all_triggers");
	ts5_log(TS5_LOGLEVEL_5, "ts5_write_all_triggers(%s)\n", filename);

	FILE *fp;
	fp = fopen(filename, "a+");
	if (!fp) {
		ts5_fatal("ts5_write_all_triggers: could not open output file %s\n", 
					filename);
	}

	ALLEGRO_EVENT event;

	while (al_get_next_event(_ts5_data.timer.trigger_log, &event)) {

		fprintf(fp, "%10.8f ", event.any.timestamp);

		double timing_error = ((double)event.user.data3)/1000000.0;
		fprintf(fp, "%10.8f ", timing_error);

		if (event.type==TS5_EVENT_PARPORT_TRIGGER) {
			fprintf(fp, "PARPORT    ");
		}
		else if (event.type==TS5_EVENT_SERIALPORT_TRIGGER) {
			fprintf(fp, "SERIALPORT ");
		}
		else if (event.type==TS5_EVENT_CEDRUSBOX_TRIGGER) {
			fprintf(fp, "CEDRUSBOX  ");
		}	
		else {
			fprintf(fp, "UNKNOWN    ");
		}

		fprintf(fp, "%d ", (int)event.user.data1+1);
		fprintf(fp, "%d ", (int)event.user.data2);
		fprintf(fp, "\n");
	}

	fclose(fp);
}


////////////////////////////////////////////////////////////////////////////////
//@}
////////////////////////////////////////////////////////////////////////////////




